from __future__ import annotations

import os
from typing import Callable, Generic, Iterator, TypeVar

from .default import Default
from .panic import Panic
from .utils import NULL

__all__ = ["Result"]

E = TypeVar("E")
F = TypeVar("F")

T = TypeVar("T")
U = TypeVar("U")
R = TypeVar("R")


class Result(Generic[T, E]):
    """
    Class containing either an `Ok(T)` or `Err(E)` variant

    Used for expressing where a process may be erraneous or may fail
    """

    vars()["__RAND_LEN"] = 64
    vars()["__INNER_OK_VAL"] = (
        os.urandom(vars()["__RAND_LEN"])
        .decode(errors="ignore")
        .center(vars()["__RAND_LEN"] + 2)
    )
    vars()["__INNER_ERR_VAL"] = (
        os.urandom(vars()["__RAND_LEN"])
        .decode(errors="ignore")
        .center(vars()["__RAND_LEN"] + 2)
    )
    vars()["__IS_OK"] = (
        os.urandom(vars()["__RAND_LEN"])
        .decode(errors="ignore")
        .center(vars()["__RAND_LEN"] + 2)
    )

    def __repr__(self) -> str:
        if self.is_ok():
            return f"Ok({self.unwrap()})"
        else:
            return f"Err({self.unwrap_err()})"

    def __str__(self) -> str:
        return self.__repr__()

    def __setattr__(self, _name, _value) -> None:
        pass

    # defaults

    @classmethod
    def ok(cls, val: T) -> Result[T, E]:
        """Creates a new `Ok` variant of `Result`"""
        r = cls()

        object.__setattr__(r, object.__getattribute__(cls, "__IS_OK"), True)
        object.__setattr__(r, object.__getattribute__(cls, "__INNER_OK_VAL"), val)
        object.__setattr__(r, object.__getattribute__(cls, "__INNER_ERR_VAL"), NULL)

        return r

    @classmethod
    def err(cls, val: E) -> Result[T, E]:
        """Creates a new `Err` variant of `Result`"""
        r = cls()

        object.__setattr__(r, object.__getattribute__(cls, "__IS_OK"), False)
        object.__setattr__(r, object.__getattribute__(cls, "__INNER_OK_VAL"), NULL)
        object.__setattr__(r, object.__getattribute__(cls, "__INNER_ERR_VAL"), val)

        return r

    # ok or err

    def is_ok(self) -> bool:
        """Returns if self is an `Ok` variant"""
        return not not object.__getattribute__(
            self, object.__getattribute__(self, "__IS_OK")
        )

    def is_ok_and(self, f: Callable[[T], bool]) -> bool:
        """Returns `True` if self is an `Ok` variant and matches predicate"""
        if self.is_err():
            return False

        return f(self.unwrap())

    def is_err(self) -> bool:
        """Returns `True` if self is an `Err` variant"""
        return not object.__getattribute__(
            self, object.__getattribute__(self, "__IS_OK")
        )

    def is_err_and(self, f: Callable[[E], bool]) -> bool:
        """Returns `False` if self is an `Err` variant and matches predicate"""
        if self.is_ok():
            return False

        return f(self.unwrap_err())

    # unwrap

    def unwrap(self) -> T:
        """
        If self is `Ok` variant, returns contained value

        If self is `Err` variant, panics
        """
        if self.is_err():
            raise Panic("Called `Result.unwrap` on an `Err` variant")
        else:
            return object.__getattribute__(
                self, object.__getattribute__(self, "__INNER_OK_VAL")
            )

    def unwrap_or(self, val: T) -> T:
        """
        If self is `Ok` variant, returns contained value

        If self is `Err` variant, returns specified value
        """
        if self.is_err():
            return val
        else:
            return self.unwrap()

    def unwrap_or_else(self, f: Callable[[E], T]) -> T:
        """
        If self is `Ok` variant, returns contained value

        If self is `Err` variant, returns predicate result, passing it the err
        """
        if self.is_err():
            return f(self.unwrap_err())
        else:
            return self.unwrap()

    def unwrap_or_default(self, default: Default[T]) -> T:
        """
        If self is `Ok` variant, returns contained value

        If self is `Err` variant, returns default value for type
        """
        if self.is_err():
            return default.default()
        else:
            return self.unwrap()

    def unwrap_err(self) -> E:
        """
        If self is `Err` variant, returns contained value

        If self is `Ok` variant, panics
        """
        if self.is_ok():
            raise Panic("Called `Result.unwrap_err` on an `Ok` variant")
        else:
            return object.__getattribute__(
                self, object.__getattribute__(self, "__INNER_ERR_VAL")
            )

    def expect(self, msg: str) -> T:
        """
        If self is `Ok` variant, returns the contained value

        If self is `Err` variant, panics with the specified error message
        """
        if self.is_err():
            raise Panic(msg)
        else:
            return self.unwrap()

    def expect_err(self, msg: str) -> E:
        """
        If self is `Err` variant, returns the contained value

        If self is `Ok` variant, panics with the specified error message
        """
        if self.is_ok():
            raise Panic(msg)
        else:
            return self.unwrap_err()

    # contains

    def contains(self, val: U) -> bool:
        """Returns whether the contained `Ok` value matches specified value"""
        if self.is_ok():
            return self.unwrap() == val
        else:
            return False

    def contains_err(self, val: F) -> bool:
        """Returns whether the contained `Err` value matches specified value"""
        if self.is_err():
            return self.unwrap_err() == val
        else:
            return False

    # map

    def map(self, f: Callable[[T], U]) -> Result[U, E]:
        """If self is an `Ok` variant, transform it with the predicate"""
        if self.is_ok():
            return Result.ok(f(self.unwrap()))
        else:
            return Result.err(self.unwrap_err())

    def map_or(self, default: U, f: Callable[[T], U]) -> U:
        """
        If self is an `Ok` variant, transform it with the predicate

        If self is an `Err` variant, replace it with the value
        """
        if self.is_ok():
            return f(self.unwrap())
        else:
            return default

    def map_or_else(self, default: Callable[[E], U], f: Callable[[T], U]) -> U:
        """
        If self is an `Ok` variant, transform it with the predicate

        If self is an `Err` variant, replace it with the predicate value
        """
        if self.is_ok():
            return f(self.unwrap())
        else:
            return default(self.unwrap_err())

    def map_err(self, f: Callable[[E], F]) -> Result[T, F]:
        """If self is an `Err` variant, transform it with the predicate f"""
        if self.is_ok():
            return Result.ok(self.unwrap())
        else:
            return Result.err(f(self.unwrap_err()))

    # inspect

    def inspect(self, f: Callable[[T], None]) -> Result[T, E]:
        """Calls a predicate on an `Ok` variant without modifying it"""
        if self.is_ok():
            f(self.unwrap())

            return Result.ok(self.unwrap())
        else:
            return Result.err(self.unwrap_err())

    def inspect_err(self, f: Callable[[E], None]) -> Result[T, E]:
        """Calls a predicate on an `Err` variant without modifying the value"""
        if self.is_ok():
            return Result.ok(self.unwrap())
        else:
            f(self.unwrap_err())

            return Result.err(self.unwrap_err())

    # iter

    def iter(self) -> Iterator[T]:
        """Transforms self into an iterator containing the `Ok` variant"""
        if self.is_ok():
            yield self.unwrap()

        return

    def iter_err(self) -> Iterator[E]:
        """Transforms self into an iterator containing the `Err` variant"""
        if self.is_err():
            yield self.unwrap_err()

        return



from oxypy import Option, Result


def test_option_variant_validation():
    none_opt = Option.none()
    some_opt = Option.some(10)

    assert none_opt.is_none() is True
    assert none_opt.is_some() is False

    assert some_opt.is_none() is False
    assert some_opt.is_some() is True


def test_result_variant_validation():
    ok_res = Result.ok(10)
    err_res = Result.err(20)

    assert ok_res.is_ok() is True
    assert ok_res.is_err() is False

    assert err_res.is_ok() is False
    assert err_res.is_err() is True


def test_some_option_unwrapping():
    some_opt = Option.some(100)

    assert some_opt.unwrap() == 100



import sys
import typing


class Panic(Exception):
    """
    A very basic Exception wrapper for Rust's `panic!()` functionality
    """
    def __init__(self, message: str) -> None:
        self.__message = message
        super().__init__()

    def __repr__(self) -> str:
        name = type(self).__name__
        message = self.__message

        return f"{name}({message!r})"

    def __call__(*args, **kwargs) -> typing.NoReturn:
        sys.exit(1)


def panic(*, msg: str) -> typing.NoReturn:
    print("Panicked:", msg)
    sys.exit(1)



"""

"""

__all__ = ["Either", "Default", "Option", "Panic", "Result"]
__version__ = "0.1.0"
__version_info__ = (0, 1, 0)
__author__ = "Ross Morgan"
__email__ = "rmorgan512@protonmail.ch"

from .default import Default
from .panic import Panic

from .either import Either
from .option import Option
from .result import Result



from typing import Generic, Protocol, Type, TypeVar

T_co = TypeVar("T_co", covariant=True)


class Default(Protocol, Generic[T_co]):
    """
    A protocol that gives a type to ability to have a default value
    """
    @classmethod
    def default(cls: Type[T_co]) -> T_co:
        ...  # noqa



from __future__ import annotations

import os
from typing import Callable, Generic, TypeVar

from .default import Default
from .panic import Panic
from .utils import NULL

__all__ = ["Either"]

L = TypeVar("L")
R = TypeVar("R")

C = TypeVar("C")
T = TypeVar("T")

K = TypeVar("K")
S = TypeVar("S")


class Either(Generic[L, R]):
    """
    Class containing either an `Left(L)` or `Right(R)` variant

    Used for expressing where a value may be one of two types
    """

    vars()["__RAND_LEN"] = 64
    vars()["__IS_LEFT"] = (
        os.urandom(vars()["__RAND_LEN"])
        .decode(errors="ignore")
        .center(vars()["__RAND_LEN"] + 2)
    )
    vars()["__INNER_LEFT_VAL"] = (
        os.urandom(vars()["__RAND_LEN"])
        .decode(errors="ignore")
        .center(vars()["__RAND_LEN"] + 2)
    )
    vars()["__INNER_RIGHT_VAL"] = (
        os.urandom(vars()["__RAND_LEN"])
        .decode(errors="ignore")
        .center(vars()["__RAND_LEN"] + 2)
    )

    def __repr__(self) -> str:
        if self.is_left():
            return f"Left({self.unwrap_left()})"
        else:
            return f"Right({self.unwrap_right()})"

    def __str__(self) -> str:
        return self.__repr__()

    def __setattr__(self, _name, _value) -> None:
        raise NotImplementedError

    # defaults

    @classmethod
    def left(cls, val: L) -> Either[L, R]:
        """Creates a new `Left` variant of `Either`"""
        r = cls()

        object.__setattr__(r, object.__getattribute__(cls, "__IS_LEFT"), True)
        object.__setattr__(r, object.__getattribute__(cls, "__INNER_LEFT_VAL"), val)
        object.__setattr__(r, object.__getattribute__(cls, "__INNER_RIGHT_VAL"), NULL)

        return r

    @classmethod
    def right(cls, val: R) -> Either[L, R]:
        """Creates a new `Right` variant of `Either`"""
        r = cls()

        object.__setattr__(r, object.__getattribute__(cls, "__IS_LEFT"), False)
        object.__setattr__(r, object.__getattribute__(cls, "__INNER_LEFT_VAL"), NULL)
        object.__setattr__(r, object.__getattribute__(cls, "__INNER_RIGHT_VAL"), val)

        return r

    # ok or err

    def is_left(self) -> bool:
        """Returns `True` if self is `Left` variant"""
        return not not object.__getattribute__(
            self, object.__getattribute__(self, "__IS_LEFT")
        )

    def is_right(self) -> bool:
        """Returns `True` if self is `Right` variant"""
        return not object.__getattribute__(
            self, object.__getattribute__(self, "__IS_LEFT")
        )

    # unwrap

    def unwrap_left(self) -> L:
        """
        If self is `Left` variant, returns contained value

        If self is `Right` variant, panics
        """
        if self.is_right():
            raise Panic("Called `Either.unwrap_left` on a `Right` variant")
        else:
            return object.__getattribute__(
                self, object.__getattribute__(self, "__INNER_LEFT_VAL")
            )

    def unwrap_right(self) -> R:
        """
        If self is `Left` variant, panics

        If self is `Right` variant, returns contained value
        """
        if self.is_left():
            raise Panic("Called `Either.unwrap_right` on a `Left` variant")
        else:
            return object.__getattribute__(
                self, object.__getattribute__(self, "__INNER_RIGHT_VAL")
            )

    # expect

    def expect_left(self, msg: str) -> L:
        """
        If self is `Left` variant, returns the contained value

        If self is `Right` variant, panics with the specified error message
        """
        if self.is_right():
            raise Panic(msg)
        else:
            return self.unwrap_left()

    def expect_right(self, msg: str) -> R:
        """
        If self is `Left` variant, panics with the specified error message

        If self is `Right` variant, returns the contained value
        """
        if self.is_left():
            raise Panic(msg)
        else:
            return self.unwrap_right()

    # either

    def either(self, f: Callable[[L], T], g: Callable[[R], T]) -> T:
        """
        If self is `Left` variant, return the contained value wrapped in `f`

        If self is `Right` variant, returns the contained value wrapped in `g`
        """
        if self.is_left():
            return f(self.unwrap_left())
        else:
            return g(self.unwrap_right())

    def either_with(self, ctx: C, f: Callable[[C, L], T], g: Callable[[C, R], T]) -> T:
        """
        If self is `Left` variant, return the contained value wrapped in `f`

        If self is `Right` variant, returns the contained value wrapped in `g`

        Like `Either.either` but supplies extra context to which function
        ends up being called
        """
        if self.is_left():
            return f(ctx, self.unwrap_left())
        else:
            return g(ctx, self.unwrap_right())

    # and then

    def left_and_then(self, f: Callable[[L], K]) -> Either[K, R]:
        """
        If self is `Left` variant, wraps the value in `f`

        If self is `Right` variant, returns a copy of self
        """
        inner: Either[K, R]

        if self.is_left():
            inner = Either.left(f(self.unwrap_left()))
        else:
            inner = Either.right(self.unwrap_right())

        return inner

    def right_and_then(self, g: Callable[[R], S]) -> Either[L, S]:
        """
        If self is `Left` variant, returns a copy of self

        If self is `Right` variant, wraps the value in `f`
        """
        inner: Either[L, S]

        if self.is_left():
            inner = Either.left(self.unwrap_left())
        else:
            inner = Either.right(g(self.unwrap_right()))

        return inner

    # or

    def left_or(self, other: L) -> L:
        """
        If self is `Left` variant, returns contained value

        If self is `Right` variant, returns given value
        """
        if self.is_left():
            return self.unwrap_left()
        else:
            return other

    def left_or_default(self, default: Default[L]) -> L:
        """
        If self is `Left` variant, returns contained value

        If self is `Right` variant, returns type's default value
        """
        if self.is_left():
            return self.unwrap_left()
        else:
            return default.default()

    def left_or_else(self, g: Callable[[R], L]) -> L:
        """
        If self is `Left` variant, returns contained value

        If self is `Right` variant, returns result of given function `f`
        """
        if self.is_left():
            return self.unwrap_left()
        else:
            return g(self.unwrap_right())

    def right_or(self, other: R) -> R:
        """
        If self is `Left` variant, returns given value

        If self is `Right` variant, returns contained value
        """
        if self.is_right():
            return self.unwrap_right()
        else:
            return other

    def right_or_default(self, default: Default[R]) -> R:
        """
        If self is `Left` variant, returns type's default value

        If self is `Right` variant, returns contained value
        """
        if self.is_right():
            return self.unwrap_right()
        else:
            return default.default()

    def right_or_else(self, f: Callable[[L], R]) -> R:
        """
        If self is `Left` variant, returns result of given function `f`

        If self is `Right` variant, returns contained value
        """
        if self.is_right():
            return self.unwrap_right()
        else:
            return f(self.unwrap_left())
